@@ -109,7 +109,7 @@ use Time::HiRes ();
 my $opt_bsd_resource = eval "use BSD::Resource; 1;";
 
 use vars qw{$VERSION};
-$VERSION = "1.45";
+$VERSION = "1.46";
 
 use warnings;
 no  warnings qw(deprecated);
@@ -400,7 +400,7 @@ sub EpollEventLoop {
 
     foreach my $fd ( keys %OtherFds ) {
         if (epoll_ctl($Epoll, EPOLL_CTL_ADD, $fd, EPOLLIN) == -1) {
-            print STDERR "epoll_ctl(): failure adding fd=$fd; $! (", $!+0, ")\n";
+            warn "epoll_ctl(): failure adding fd=$fd; $! (", $!+0, ")\n";
         }
     }
 
@@ -430,7 +430,7 @@ sub EpollEventLoop {
                     $code->($state);
                 } else {
                     my $fd = $ev->[0];
-                    print STDERR "epoll() returned fd $fd w/ state $state for which we have no mapping.  removing.\n";
+                    warn "epoll() returned fd $fd w/ state $state for which we have no mapping.  removing.\n";
                     POSIX::close($fd);
                     epoll_ctl($Epoll, EPOLL_CTL_DEL, $fd, 0);
                 }
@@ -575,7 +575,7 @@ sub KQueueEventLoop {
                 if (my $code = $OtherFds{$fd}) {
                     $code->($filter);
                 }  else {
-                    print STDERR "kevent() returned fd $fd for which we have no mapping.  removing.\n";
+                    warn "kevent() returned fd $fd for which we have no mapping.  removing.\n";
                     POSIX::close($fd); # close deletes the kevent entry
                 }
                 next;
@@ -643,7 +643,17 @@ sub PostEventLoop {
     # now we can close sockets that wanted to close during our event processing.
     # (we didn't want to close them during the loop, as we didn't want fd numbers
     #  being reused and confused during the event loop)
-    $_->close while ($_ = shift @ToClose);
+    while (my $sock = shift @ToClose) {
+        my $fd = fileno($sock);
+
+        # close the socket.  (not a Danga::Socket close)
+        $sock->close;
+
+        # and now we can finally remove the fd from the map.  see
+        # comment above in _cleanup.
+        delete $DescriptorMap{$fd};
+    }
+
 
     # by default we keep running, unless a postloop callback (either per-object
     # or global) cancels it
@@ -677,6 +687,10 @@ sub new {
 
     $self->{sock}        = $sock;
     my $fd = fileno($sock);
+
+    Carp::cluck("undef sock and/or fd in Danga::Socket->new.  sock=" . ($sock || "") . ", fd=" . ($fd || ""))
+        unless $sock && $fd;
+
     $self->{fd}          = $fd;
     $self->{write_buf}      = [];
     $self->{write_buf_offset} = 0;
@@ -701,6 +715,9 @@ sub new {
                         IO::KQueue::EV_ADD() | IO::KQueue::EV_DISABLE());
     }
 
+    Carp::cluck("Danga::Socket::new blowing away existing descriptor map for fd=$fd ($DescriptorMap{$fd})")
+        if $DescriptorMap{$fd};
+
     $DescriptorMap{$fd} = $self;
     return $self;
 }
@@ -774,7 +791,7 @@ sub close {
     if (DebugLevel) {
         my ($pkg, $filename, $line) = caller;
         my $reason = $_[1] || "";
-        print STDERR "Closing \#$self->{fd} due to $pkg/$filename/$line ($reason)\n";
+        warn "Closing \#$self->{fd} due to $pkg/$filename/$line ($reason)\n";
     }
 
     # this does most of the work of closing us
@@ -818,10 +835,18 @@ sub _cleanup {
 
     # now delete from mappings.  this fd no longer belongs to us, so we don't want
     # to get alerts for it if it becomes writable/readable/etc.
-    delete $DescriptorMap{$self->{fd}};
     delete $PushBackSet{$self->{fd}};
     delete $PLCMap{$self->{fd}};
 
+    # we explicitly don't delete from DescriptorMap here until we
+    # actually close the socket, as we might be in the middle of
+    # processing an epoll_wait/etc that returned hundreds of fds, one
+    # of which is not yet processed and is what we're closing.  if we
+    # keep it in DescriptorMap, then the event harnesses can just
+    # looked at $pob->{closed} and ignore it.  but if it's an
+    # un-accounted for fd, then it (understandably) freak out a bit
+    # and emit warnings, thinking their state got off.
+
     # and finally get rid of our fd so we can't use it anywhere else
     $self->{fd} = undef;
 }
@@ -1095,9 +1120,9 @@ sub dump_error {
         push @list, "\t$file:$line called $sub\n";
     }
 
-    print STDERR "ERROR: $_[1]\n" .
-                 "\t$_[0] = " . $_[0]->as_string . "\n" .
-                 join('', @list);
+    warn "ERROR: $_[1]\n" .
+        "\t$_[0] = " . $_[0]->as_string . "\n" .
+        join('', @list);
 }
 
 
